from __future__ import absolute_import
import logging
import json
from ..codes import errorcodes
from ..codes import opcodes
from ..codes import statecodes
from ..util.utils import is_callable, json2dict, range
from .ipc import BaseConnection


_RPC_VERSION = 1


class RpcConnection(object):
    _connection = None
    _state = statecodes.Disconnected
    _app_id = None
    _last_err_code = 0
    _last_err_msg = ''
    _pipe_no = 0
    _on_connect = None
    _on_disconnect = None

    def __init__(self, app_id, pipe_no=0, log=True, logger=None, log_file=None, log_level=logging.INFO):
        self._connection = BaseConnection(log=log, logger=logger, log_file=log_file, log_level=log_level)
        self._app_id = str(app_id)
        if pipe_no in range(0, 10):
            self._pipe_no = pipe_no

    def open(self):
        if self.state == statecodes.Connected:
            self.log('debug', 'Already connected; no need to open.')
            return

        if self.state == statecodes.Disconnected:
            self.connection.open(pipe_no=self._pipe_no)
            if not self.connection.is_open:
                self.log('warning', 'Failed to open IPC connection.')
                return

        if self.state == statecodes.SentHandshake:
            did_read, data = self.read()
            if did_read:
                cmd = data.get('cmd', None)
                evt = data.get('evt', None)
                if all(x is not None for x in (cmd, evt)) and cmd == 'DISPATCH' and evt == 'READY':
                    self.state = statecodes.Connected
                    if self.on_connect is not None:
                        self.on_connect(data)
                    self.log('info', 'IPC connected successfully.')
        else:
            data = {'v': _RPC_VERSION, 'client_id': self.app_id}
            if self.connection.write(json.dumps(data), opcodes.Handshake):
                self.state = statecodes.SentHandshake
                self.log('debug', 'IPC connection sent handshake.')
            else:
                self.log('warning', 'IPC failed to send handshake.')
                self.close()

    def close(self):
        if self.on_disconnect is not None and self.state in (statecodes.Connected, statecodes.SentHandshake):
            self.on_disconnect(self._last_err_code, self._last_err_msg)
            self.log('debug', 'Attempting to close IPC connection.')
        if self.connection is not None:
            self.connection.close()
        else:
            self.log('warning', 'Called close without a connection!')
        self.state = statecodes.Disconnected

    def write(self, data):
        if isinstance(data, dict):
            data = json.dumps(data)
        if not self.connection.write(data, opcodes.Frame):
            self.log('warning', 'Failed to write frame to IPC connection.')
            self.close()
            return False
        return True

    def read(self):
        if self.state not in (statecodes.Connected, statecodes.SentHandshake):
            self.log('debug', 'We aren\'t connected, therefore we cannot read data yet.')
            return False
        while True:
            did_read, opcode, data = self.connection.read()
            self.log('debug', 'ipc.read(): read: {}, Opcode: {}, data: {}'.format(did_read, opcode, data))
            if not did_read:
                err_reason = data[0]
                if (err_reason == errorcodes.PipeClosed and not self.connection.is_open) \
                        or err_reason == errorcodes.ReadCorrupt:
                    self._last_err_code = err_reason
                    self._last_err_msg = data[1]
                    self.log('debug', 'Failed to read; Connection closed. {}'.format(data))
                    self.close()
                return False, None
            if opcode == opcodes.Close:
                data = json2dict(data)
                self._last_err_code = data.get('code', -1)
                self._last_err_msg = data.get('message', '')
                self.log('debug', 'Opcode == Close. Closing connection.')
                self.close()
                return False, None
            elif opcode == opcodes.Frame:
                data = json2dict(data)
                self.log('debug', 'Successful read: {}'.format(data))
                return True, data
            elif opcode == opcodes.Ping:
                if not self.connection.write('', opcodes.Pong):
                    self.log('warning', 'Failed to send Pong message.')
                    self.close()
            elif opcode == opcodes.Pong:
                # Discord does nothing here
                pass
            else:
                # something bad happened
                self._last_err_code = errorcodes.ReadCorrupt
                self._last_err_msg = 'Bad IPC frame.'
                self.log('warning', 'Got a bad frame from IPC connection.')
                self.close()
                return False, None

    def destroy(self):
        self.log('info', 'Destroying RPC connection.')
        self.close()
        self.connection.destroy()
        self._connection = None

    def log(self, *args):
        if self._connection is not None:
            self._connection.log(*args)

    @property
    def connection(self):
        return self._connection

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state):
        if isinstance(state, int) and state in (statecodes.Connected, statecodes.SentHandshake,
                                                statecodes.Disconnected, statecodes.AwaitingResponse):
            self._state = state
        else:
            self.log('warning', 'Invalid state number!')

    @property
    def app_id(self):
        return self._app_id

    @property
    def is_open(self):
        return self.state == statecodes.Connected

    @property
    def on_connect(self):
        return self._on_connect

    @on_connect.setter
    def on_connect(self, callback):
        if callback is None or is_callable(callback):
            self._on_connect = callback
        else:
            self.log('warning', 'on_connect must be callable/None!')

    @property
    def on_disconnect(self):
        return self._on_disconnect

    @on_disconnect.setter
    def on_disconnect(self, callback):
        if callback is None or is_callable(callback):
            self._on_disconnect = callback
        else:
            self.log('warning', 'on_disconnect must be callable/None!')



from __future__ import absolute_import
from .limits import INT32_MIN, INT32_MAX, UINT32_MAX, INT64_MIN, INT64_MAX, UINT64_MAX


_number_types = [int]
try:
    _number_types.append(float)
except NameError:
    pass
try:
    _number_types.append(long)
except NameError:
    pass
_number_types = tuple(_number_types)


class UnderflowError(ArithmeticError):
    pass


class Number(object):
    _min = None
    _max = None
    _bits = None
    _raise_exceptions = False

    def __init__(self, number=0, raise_exceptions=False):
        self._raise_exceptions = raise_exceptions
        if isinstance(number, Number):
            number = number.get_number()
        else:
            # validate number, ret_val should only be number, not a class
            number = self._check_number(number, number_only=True)
        if not isinstance(number, _number_types):
            raise TypeError('Number must be of type int/float/long!')
        self._number = number

    def _check_number(self, num, number_only=False):
        if self._max is not None and self._min is not None:
            if self._raise_exceptions:
                if num > self._max:
                    raise OverflowError()
                elif num < self._min:
                    raise UnderflowError()
            if self._min == 0:
                if not number_only:
                    return self.__class__(num & self._max)
                else:
                    return num & self._max
            else:
                if num & (1 << (self._bits-1)):
                    if not number_only:
                        return self.__class__(num | ~self._max)
                    else:
                        return num | ~self._max
                else:
                    if not number_only:
                        return self.__class__(num & self._max)
                    else:
                        return num & self._max
        else:
            # there are no min/max to check with, so just return the number
            if not number_only:
                return self.__class__(num)
            else:
                return num

    def __lt__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._number < other

    def __le__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._number <= other

    def __eq__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._number == other

    def __ne__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._number != other

    def __gt__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._number > other

    def __ge__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._number >= other

    def __bool__(self):
        return self._number

    def __repr__(self):
        return "{class_name}({number})".format(class_name=self.__class__.__name__, number=self._number)

    def __str__(self):
        return str(self._number)

    def __cmp__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        if self._number == other:
            return 0
        elif self._number < other:
            return -1
        elif self._number > other:
            return 1

    def __nonzero__(self):
        return self._number != 0

    def __add__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(self._number + other)

    def __sub__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(self._number - other)

    def __mul__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(self._number * other)

    def __div__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(self._number / other)

    def __truediv__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self.__div__(other)

    def __floordiv__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(self._number // other)

    def __mod__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(self._number % other)

    def __divmod__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(divmod(self._number, other))

    def __pow__(self, power, modulo=None):
        if isinstance(power, Number):
            power = power.get_number()
        if modulo is not None and isinstance(modulo, Number):
            modulo = modulo.get_number()
        return self._check_number(pow(self._number, power, modulo))

    def __lshift__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(self._number << other)

    def __rshift__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(self._number >> other)

    def __and__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(self._number & other)

    def __xor__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(self._number ^ other)

    def __or__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(self._number | other)

    def __radd__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(other + self._number)

    def __rsub__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(other - self._number)

    def __rmul__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(other * self._number)

    def __rdiv__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(other / self._number)

    def __rtruediv__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self.__rdiv__(other)

    def __rfloordiv__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(other // self._number)

    def __rmod__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(other % self._number)

    def __rdivmod__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(divmod(other, self._number))

    def __rpow__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(pow(other, self._number))

    def __rlshift__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(other << self._number)

    def __rrshift__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(other >> self._number)

    def __rand__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(other & self._number)

    def __rxor__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(other ^ self._number)

    def __ror__(self, other):
        if isinstance(other, Number):
            other = other.get_number()
        return self._check_number(other | self._number)

    def __neg__(self):
        return self._check_number(-self._number)

    def __pos__(self):
        return self._check_number(+self._number)

    def __abs__(self):
        return self._check_number(abs(self._number))

    def __invert__(self):
        return self._check_number(~self._number)

    def __complex__(self):
        return complex(self._number)

    def __int__(self):
        return int(self._number)

    def __long__(self):
        return long(self._number)

    def __float__(self):
        return float(self._number)

    def __oct__(self):
        return oct(self._number)

    def __hex__(self):
        return hex(self._number)

    def __round__(self, ndigits=None):
        if ndigits is not None and isinstance(ndigits, Number):
            ndigits = ndigits.get_number()
        return round(self._number, ndigits)

    def __trunc__(self):
        from math import trunc
        return trunc(self._number)

    def __floor__(self):
        from math import floor
        return floor(self._number)

    def __ceil__(self):
        from math import ceil
        return ceil(self._number)

    def get_number(self):
        return self._number

    def get_copy(self):
        return self.__class__(self._number)


class _Int(Number):
    def __init__(self, number=0, raise_exceptions=False):
        if isinstance(number, Number):
            number = number.get_number()
        number = int(number)
        Number.__init__(self, number, raise_exceptions)

    def _check_number(self, num, number_only=False):
        if isinstance(num, Number):
            num = num.get_number()
        num = int(num)
        return Number._check_number(self, num, number_only)


class Int32(_Int):
    _min = INT32_MIN
    _max = INT32_MAX
    _bits = 32


class Int64(_Int):
    _min = INT64_MIN
    _max = INT64_MAX
    _bits = 64


class UInt32(_Int):
    _min = 0
    _max = UINT32_MAX


class UInt64(_Int):
    _min = 0
    _max = UINT64_MAX
